from __future__ import annotations
import typing as t
import datetime as dt
import dateparser
import pydantic

from spice_rack import _logging
from spice_rack._ts_service._tz_key import TimeZoneKey


__all__ = (
    "Timestamp",
)


_TzKeyT = t.Union[str, TimeZoneKey, t.Literal["local"]]
_PythonTimestampT = float  # seconds from epoch with decimals as sub-second measurements


class Timestamp(pydantic.RootModel[int], _logging.log_extra.LoggableObjMixin):
    """
    special subclass of 'int' that contains the utc millisecond from epoch.

    This class will automatically validate date and datetime objects and ints. Use extended type annotations
    for more flexible parsing.

    """
    _default_assumed_tz: t.ClassVar[TimeZoneKey] = TimeZoneKey("UTC")
    """timezone we assume when parsing raw data into this object"""
    # _default_tz: ClassVar[TimeZoneKey] = TimeZoneKey.local()

    def to_python_timestamp(self) -> _PythonTimestampT:
        # convert to seconds from epoch, posix timestamp
        return float(self.root) / 1000

    def to_dt_obj(
            self,
            with_tz: t.Optional[_TzKeyT] = None
    ) -> dt.datetime:
        res: dt.datetime

        if with_tz:
            utc_dt_obj = dt.datetime.fromtimestamp(
                self.to_python_timestamp(), tz=TimeZoneKey("UTC").as_zone_info()
            )
            tz_key: TimeZoneKey
            if with_tz == "local":
                tz_key = TimeZoneKey.local()
            else:
                tz_key = TimeZoneKey(str(with_tz))

            new_tz_info = tz_key.as_zone_info()
            obj_new_tz = utc_dt_obj.astimezone(tz=new_tz_info)
            res = obj_new_tz
        else:
            res = dt.datetime.fromtimestamp(
                self.to_python_timestamp(), tz=None
            )
        return res

    def to_iso_str(
            self,
            with_tz: t.Optional[_TzKeyT] = None
    ) -> str:
        """
        returns the datetime object as a formatted str

        Args:
            with_tz: a timezone key we can specify. use 'local' to use whatever timezone is local
                to the process calling this function
        Returns:
            str: formatted string
        """
        dt_obj = self.to_dt_obj(with_tz=with_tz)
        return dt_obj.isoformat()

    def to_pretty_str(
            self,
            fmat: str = "%x %X %Z",
            with_tz: t.Optional[_TzKeyT] = "local"
    ) -> str:
        """
        returns the datetime object as a formatted str, in a pretty, super readable format.
        The default is "%c %Z"

        Args:
            fmat: Optional custom formatting. See python datetime docs for formatting help.
                If not specified, we use 'auto' selected appropriate format.

            with_tz: a timezone key we can specify. use 'local' to use whatever timezone is local
                to the process calling this function
        Returns:
            str: formatted string

        Notes:
            - see the docs on format codes:
              https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
        """
        dt_obj = self.to_dt_obj(with_tz=with_tz)
        return dt_obj.strftime(fmat)

    def special_repr(self, with_tz: t.Optional[_TzKeyT] = None) -> str:
        return f"{self.__class__.__name__}['{self.to_iso_str(with_tz=with_tz)}']"

    def to_file_path_fmat(self) -> str:
        return self.to_dt_obj().strftime("%Y_%m_%dT%H_%M_%S_%f")

    @classmethod
    def _from_datetime_obj(
            cls,
            value: dt.datetime,
            assumed_tz: TimeZoneKey
    ) -> int:
        tz_info: TimeZoneKey
        if value.tzinfo:
            tz_info = TimeZoneKey(value.tzname())
        else:
            tz_info = assumed_tz

        dt_obj_utc = value.astimezone(tz_info.as_zone_info())
        return int(dt_obj_utc.timestamp() * 1000)

    @classmethod
    def _from_date_obj(
            cls,
            value: dt.date,
            assumed_tz: TimeZoneKey
    ) -> int:
        datetime_obj = dt.datetime.fromisoformat(value.isoformat())
        return cls._from_datetime_obj(datetime_obj, assumed_tz=assumed_tz)

    @classmethod
    def _from_str(
            cls,
            value: str,
            assumed_tz: TimeZoneKey
    ) -> int:
        """parse a str to datetime object then to timestamp"""
        dt_obj_maybe = dateparser.parse(value)
        if dt_obj_maybe is None:
            raise ValueError(
                f"cannot parse the string, '{value}', to a datetime object "
            )

        else:
            dt_obj: dt.datetime = dt_obj_maybe
        return cls._from_datetime_obj(value=dt_obj, assumed_tz=assumed_tz)

    @classmethod
    def _from_float(
            cls,
            value: float,
            assumed_tz: TimeZoneKey
    ) -> int:
        raise ValueError("initializing from a float is not currently supported")

    @pydantic.model_validator(mode="before")
    def _parse_non_int(cls: t.Type[Timestamp], data: t.Any) -> t.Any:

        if isinstance(data, dt.datetime):
            data = cls._from_datetime_obj(data, assumed_tz=cls._default_assumed_tz)

        elif isinstance(data, dt.date):
            data = cls._from_date_obj(data, assumed_tz=cls._default_assumed_tz)

        return data

    @classmethod
    def now(cls) -> Timestamp:
        """get now down to the millisecond, not microsecond"""
        return Timestamp(dt.datetime.utcnow())

    @classmethod
    def today(cls) -> Timestamp:
        """get the current date, as a Timestamp object."""
        dt_obj = dt.datetime.today()
        return Timestamp(dt_obj)

    def __get_logger_data__(self) -> _logging.log_extra.ExtraLogData:
        return _logging.log_extra.ExtraLogData(
            key="timestamp",
            desc="a timestamp object",
            data=self.to_iso_str()
        )



from __future__ import annotations
import typing as t
import dateparser
import datetime as dt

import pydantic
import pydantic_core
from spice_rack import _bases


__all__ = (
    "StrParserException",
    "StrParser",
    "StrParserStrict"
)


class _StrParserErrorInfo(_bases.exceptions.ErrorInfoBase):
    raw_str: str = pydantic.Field(description="the string we failed to parse into date format")
    strict: bool = pydantic.Field(description="if true, this failed strict mode parsing")
    specified_date_formats: t.List[str] = pydantic.Field(description="the date formats specified ")


class StrParserException(_bases.exceptions.CustomExceptionBase[_StrParserErrorInfo]):
    def __init__(self, raw_str: str, strict: bool, specified_date_formats: t.List[str]):
        error_info = {
            "raw_str": raw_str,
            "strict": strict,
            "specified_date_formats": specified_date_formats
        }

        if strict:
            detail = f"the strict string parser failed to parse '{raw_str}'"
        else:
            detail = f"the flexible string parser failed to parse '{raw_str}'"
        super().__init__(
            detail=detail,
            error_info=error_info,
            verbose=True,
            extra_info=None
        )


class StrParser:
    """
    flexible string parser using dateparser library under the hood. If you provide date_formats it will be as
    fast as a strict parser, but fallback to the slower approach if needed.
    """
    date_formats: t.List[str]
    """specify these to indicate that the str parser should try these date formats first"""

    def __init__(self, date_formats: t.Optional[t.List[str]] = None, strict: bool = False):
        self.date_formats = date_formats if date_formats else []
        self.strict = strict

    def _get_func(self) -> t.Callable[[t.Any, ], t.Any]:
        def func(data: t.Any) -> t.Any:
            if isinstance(data, str):
                parsed_maybe = dateparser.parse(data, date_formats=self.date_formats, )
                if parsed_maybe is None:
                    raise StrParserException(
                        raw_str=data,
                        specified_date_formats=self.date_formats,
                        strict=False
                    ).as_pydantic_error()
                else:
                    data = parsed_maybe
            return data
        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_before_validator_function(
            function=val_func,
            schema=schema
        )
        return schema


class StrParserStrict:
    """
    This will try the different string formats in order, and choose the first one that works.
    """
    date_formats: t.List[str]
    """specify these to indicate that the str parser should try these date formats first"""

    def __init__(self, date_formats: t.List[str]):
        if len(date_formats) == 0:
            raise ValueError(
                "You must specify at least one date_format for the StrParserStrict validator"
            )
        self.date_formats = date_formats

    def _get_func(self) -> t.Callable[[t.Any, ], t.Any]:
        def func(data: t.Any) -> t.Any:
            if isinstance(data, str):
                parsed_maybe: t.Optional[dt.datetime] = None
                for format_i in self.date_formats:
                    try:
                        parsed_maybe = dt.datetime.strptime(data, format_i)
                    except ValueError:
                        continue
                    except Exception as e:
                        raise e
                if parsed_maybe is None:
                    raise StrParserException(
                        raw_str=data,
                        specified_date_formats=self.date_formats,
                        strict=True
                    ).as_pydantic_error()
                else:
                    data = parsed_maybe
            return data
        return func

    def __get_pydantic_core_schema__(
            self,
            __source: t.Any,
            __handler: pydantic.GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        val_func = self._get_func()

        # todo: add info to the schema
        schema = __handler(__source)
        schema = pydantic_core.core_schema.no_info_before_validator_function(
            function=val_func,
            schema=schema
        )
        return schema



from __future__ import annotations
from typing import Optional
import zoneinfo
import tzlocal

from spice_rack import _bases


__all__ = (
    "TimeZoneKey",
    "InvalidTimeZoneKeyException"
)


# todo: specify these or find another way
class TimeZoneKey(_bases.special_str.SpecialStrBase):
    """
    a special string representing a timezone.

    see python docs for "zoneinfo" for options:
        https://docs.python.org/3/library/zoneinfo.html
    """

    @classmethod
    def timezone_choices(cls) -> list[str]:
        return list(zoneinfo.available_timezones())

    @classmethod
    def _format_str_val(cls, root_data: str) -> str:
        # ensure it is a valid zone info
        choices = zoneinfo.available_timezones()
        if root_data not in choices:
            raise InvalidTimeZoneKeyException(
                invalid_value=root_data,
                options=list(choices)
            )

        else:
            return root_data

    def as_zone_info(self) -> zoneinfo.ZoneInfo:
        return zoneinfo.ZoneInfo(str(self))

    @classmethod
    def local(cls) -> TimeZoneKey:
        """gets timezone info for where this process is running"""
        return TimeZoneKey(tzlocal.get_localzone().key)


# class _InvalidTimeZoneKeyErrorInfo(exception_base.ErrorInfoBase):
#     invalid_value: str
#     options: list[str]
#
#
# class InvalidTimeZoneKeyException(exception_base.ExceptionBase[_InvalidTimeZoneKeyErrorInfo]):
#     def __init__(
#             self,
#             invalid_value: str,
#             options: list[str],
#             extra_info: Optional[dict] = None,
#             verbose: bool = False
#     ):
#         error_info = {
#             "invalid_value": invalid_value,
#             "options": options
#         }
#
#         detail = f"'{invalid_value}' is not a valid timezone key."
#         super().__init__(
#             detail=detail,
#             error_info=error_info,
#             extra_info=extra_info,
#             verbose=verbose
#         )


# for now
class InvalidTimeZoneKeyException(Exception):
    def __init__(
            self,
            invalid_value: str,
            options: list[str],
            extra_info: Optional[dict] = None,
            verbose: bool = False
    ):
        # error_info = {
        #     "invalid_value": invalid_value,
        #     "options": options
        # }
        detail = f"'{invalid_value}' is not a valid timezone key."
        # super().__init__(
        #     detail=detail,
        #     error_info=error_info,
        #     extra_info=extra_info,
        #     verbose=verbose
        # )
        super().__init__(detail)
